# Use standard logging in this module.
import logging

# Class managing list of things.
from dls_utilpack.things import Things

from soakdb3_api.datafaces.constants import DatafaceTypes

# Exceptions.
from soakdb3_api.exceptions import NotFound

logger = logging.getLogger(__name__)

# -----------------------------------------------------------------------------------------
__default_dataface = None


def datafaces_set_default(dataface):
    global __default_dataface
    __default_dataface = dataface


def datafaces_get_default():
    global __default_dataface
    if __default_dataface is None:
        raise RuntimeError("datafaces_get_default instance is None")
    return __default_dataface


# -----------------------------------------------------------------------------------------


class Datafaces(Things):
    """
    List of available datafaces.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, name=None):
        Things.__init__(self, name)

    # ----------------------------------------------------------------------------------------
    def build_object(self, specification):
        """"""

        dataface_class = self.lookup_class(specification["type"])

        try:
            dataface_object = dataface_class(specification)
        except Exception as exception:
            raise RuntimeError(
                "unable to build dataface object for type %s" % (dataface_class)
            ) from exception

        return dataface_object

    # ----------------------------------------------------------------------------------------
    def lookup_class(self, class_type):
        """"""

        if class_type == DatafaceTypes.AIOHTTP:
            from soakdb3_api.datafaces.aiohttp import Aiohttp

            return Aiohttp

        if class_type == DatafaceTypes.DIRECT:
            from soakdb3_lib.datafaces.aiosqlite import Aiosqlite

            return Aiosqlite

        raise NotFound("unable to get dataface class for type %s" % (class_type))



class Keywords:
    COMMAND = "soakdb3::datafaces::keywords::command"
    PAYLOAD = "soakdb3::datafaces::keywords::payload"


class Commands:
    EXECUTE = "soakdb3::datafaces::commands::execute"


class DatafaceTypes:
    AIOHTTP = "soakdb3_lib.datafaces.aiohttp"
    DIRECT = "soakdb3_lib.datafaces.aiosqlite"



import asyncio

# Use standard logging in this module.
import logging

# Base class for cli subcommands.
from soakdb3_cli.subcommands.base import ArgKeywords, Base

# Context creator.
from soakdb3_lib.datafaces.context import Context

logger = logging.getLogger()


# --------------------------------------------------------------
class Service(Base):
    """
    Start single service and keep running until ^C or remotely requested shutdown.
    """

    def __init__(self, args, mainiac):
        super().__init__(args)

    # ----------------------------------------------------------------------------------------
    def run(self):
        """ """

        # Run in asyncio event loop.
        asyncio.run(self.__run_coro())

    # ----------------------------------------------------------
    async def __run_coro(self):
        """"""

        # Load the configuration.
        multiconf = self.get_multiconf(vars(self._args))
        configuration = await multiconf.load()

        # Make a client context for the soakdb3 service.
        soakdb3_context = Context(configuration["soakdb3_dataface_specification"])

        # Open the context which starts the service process.
        async with soakdb3_context:
            # Wait for it to finish.
            await soakdb3_context.server.wait_for_shutdown()

    # ----------------------------------------------------------
    def add_arguments(parser):

        parser.add_argument(
            "--configuration",
            "-c",
            help="Configuration file.",
            type=str,
            metavar="yaml filename",
            default=None,
            dest=ArgKeywords.CONFIGURATION,
        )

        return parser



import asyncio
import logging
import multiprocessing

import pytest

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class BaseTester:
    """
    This is a base class for simplest tests.
    """

    def main(self, constants, output_directory):
        """
        This is the main program which calls the test using asyncio.
        """

        multiprocessing.current_process().name = "main"

        failure_message = None
        try:
            # Run main test in asyncio event loop.
            asyncio.run(self._main_coroutine(constants, output_directory))

        except Exception as exception:
            logger.exception(
                "unexpected exception in the test method", exc_info=exception
            )
            failure_message = str(exception)

        if failure_message is not None:
            pytest.fail(failure_message)



import logging
import multiprocessing
import threading

# Utilities.
from dls_utilpack.callsign import callsign
from dls_utilpack.explain import explain
from dls_utilpack.require import require

# Basic things.
from dls_utilpack.thing import Thing

# XchemBeDataface protocolj things.
from soakdb3_api.datafaces.constants import Commands, Keywords

# Base class for an aiohttp server.
from soakdb3_lib.base_aiohttp import BaseAiohttp

# Types of dataface.
# Global dataface.
from soakdb3_lib.datafaces.datafaces import Datafaces

logger = logging.getLogger(__name__)

thing_type = "soakdb3_lib.datafaces.aiohttp"


# ------------------------------------------------------------------------------------------
class Aiohttp(Thing, BaseAiohttp):
    """
    Object implementing remote procedure calls for dataface methods.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification=None):
        Thing.__init__(self, thing_type, specification)
        BaseAiohttp.__init__(
            self,
            specification["type_specific_tbd"]["aiohttp_specification"],
        )

        self.__actual_dataface = None

    # ----------------------------------------------------------------------------------------
    def callsign(self):
        """"""
        return "%s %s" % ("XchemBeDataface.Aiohttp", BaseAiohttp.callsign(self))

    # ----------------------------------------------------------------------------------------
    def activate_process(self):
        """"""

        try:
            multiprocessing.current_process().name = "dataface"

            self.activate_process_base()

        except Exception as exception:
            logger.exception(
                f"error while activating {callsign(self)} process", exc_info=exception
            )

    # ----------------------------------------------------------------------------------------
    def activate_thread(self, loop):
        """
        Called from inside a newly created thread.
        """

        try:
            threading.current_thread().name = "soakdb3.dataface"

            self.activate_thread_base(loop)

        except Exception as exception:
            logger.exception(
                f"error while activating {callsign(self)} thread", exc_info=exception
            )

    # ----------------------------------------------------------------------------------------
    async def activate_coro(self):
        """"""
        try:
            # No special routes, we will use protocolj dispathcing only
            route_tuples = []

            # Build a local dataface for our back-end.
            self.__actual_dataface = Datafaces().build_object(
                self.specification()["type_specific_tbd"][
                    "actual_dataface_specification"
                ]
            )

            # Get the local implementation started.
            await self.__actual_dataface.start()

            await self.activate_coro_base(route_tuples)

        except Exception:
            raise RuntimeError(f"error while activating {callsign(self)} server coro")

        finally:
            if self.__actual_dataface is not None:
                try:
                    # Disconnect our local dataface connection, i.e. the one which holds the database connection.
                    await self.__actual_dataface.disconnect()

                except Exception as exception:
                    raise RuntimeError(
                        callsign(
                            self, explain(exception, "disconnecting local dataface")
                        )
                    )

    # ----------------------------------------------------------------------------------------
    async def direct_shutdown(self):
        """"""
        try:
            # Disconnect our local dataface connection, i.e. the one which holds the database connection.
            await self.__actual_dataface.disconnect()

        except Exception as exception:
            raise RuntimeError(
                callsign(self, explain(exception, "disconnecting local dataface"))
            )

        # Let the base class stop the server listener.
        await self.base_direct_shutdown()

    # ----------------------------------------------------------------------------------------
    async def __do_actually(self, function, args, kwargs):
        """"""

        # logger.info(describe("function", function))
        # logger.info(describe("args", args))
        # logger.info(describe("kwargs", kwargs))

        function = getattr(self.__actual_dataface, function)

        response = await function(*args, **kwargs)

        return response

    # ----------------------------------------------------------------------------------------
    async def dispatch(self, request_dict, opaque):
        """"""

        command = require("request json", request_dict, Keywords.COMMAND)

        if command == Commands.EXECUTE:
            payload = require("request json", request_dict, Keywords.PAYLOAD)
            response = await self.__do_actually(
                payload["function"], payload["args"], payload["kwargs"]
            )
        else:
            raise RuntimeError("invalid command %s" % (command))

        return response



import logging
import os
import tempfile
from typing import Optional

# Configurator.
from dls_multiconf_lib.constants import ThingTypes as MulticonfThingTypes
from dls_multiconf_lib.multiconfs import Multiconfs, multiconfs_set_default

# Utilities.
from dls_utilpack.visit import get_visit_year

# Environment variables with some extra functionality.
from soakdb3_lib.envvar import Envvar

logger = logging.getLogger(__name__)


class ArgKeywords:
    CONFIGURATION = "configuration"


class Base:
    """
    Base class for femtocheck subcommands.  Handles details like configuration.
    """

    def __init__(self, args):
        self._args = args

        self.__temporary_directory = None

    # ----------------------------------------------------------------------------------------
    def get_multiconf(self, args_dict: dict):

        soakdb3_multiconf = self.build_object_from_environment(args_dict=args_dict)

        # For convenience, make a temporary directory for this test.
        self.__temporary_directory = tempfile.TemporaryDirectory()

        # Make the temporary directory available to the multiconf.
        soakdb3_multiconf.substitute(
            {"temporary_directory": self.__temporary_directory.name}
        )

        substitutions = {
            "CWD": os.getcwd(),
            "HOME": os.environ.get("HOME", "HOME"),
            "USER": os.environ.get("USER", "USER"),
            "PATH": os.environ.get("PATH", "PATH"),
            "PYTHONPATH": os.environ.get("PYTHONPATH", "PYTHONPATH"),
        }

        if hasattr(self._args, "visit") and self._args.visit != "VISIT":
            BEAMLINE = os.environ.get("BEAMLINE")
            if BEAMLINE is None:
                raise RuntimeError("BEAMLINE environment variable is not defined")
            year = get_visit_year(BEAMLINE, self._args.visit)
            substitutions["BEAMLINE"] = BEAMLINE
            substitutions["VISIT"] = self._args.visit
            substitutions["YEAR"] = year

        soakdb3_multiconf.substitute(substitutions)

        # Set this as the default multiconf so it is available everywhere.
        multiconfs_set_default(soakdb3_multiconf)

        return soakdb3_multiconf

    # ----------------------------------------------------------------------------------------
    def build_object_from_environment(
        self,
        environ: Optional[dict] = None,
        args_dict: Optional[dict] = None,
    ):

        configuration_keyword = "configuration"

        multiconf_filename = None

        if args_dict is not None:
            multiconf_filename = args_dict.get(configuration_keyword)

        if multiconf_filename is not None:
            # Make sure the path exists.
            if not os.path.exists(multiconf_filename):
                raise RuntimeError(
                    f"unable to find --{configuration_keyword} file {multiconf_filename}"
                )
        else:
            # Get the explicit name of the config file.
            multiconf_filename = Envvar(
                Envvar.SOAKDB3_CONFIGFILE,
                environ=environ,
            )

            # Config file is explicitly named?
            if multiconf_filename.is_set:
                # Make sure the path exists.
                multiconf_filename = multiconf_filename.value
                if not os.path.exists(multiconf_filename):
                    raise RuntimeError(
                        f"unable to find {Envvar.SOAKDB3_CONFIGFILE} {multiconf_filename}"
                    )
            # Config file is not explicitly named?
            else:
                raise RuntimeError(
                    f"command line --{configuration_keyword} not given"
                    f" and environment variable {Envvar.SOAKDB3_CONFIGFILE} is not set"
                )

        multiconf = Multiconfs().build_object(
            {
                "type": MulticonfThingTypes.YAML,
                "type_specific_tbd": {"filename": multiconf_filename},
            }
        )

        return multiconf



import asyncio
import logging
import multiprocessing

import pytest

from soakdb3_api.databases.constants import BodyFieldnames, HeadFieldnames, Tablenames
from soakdb3_lib.databases.databases import Databases

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestDatabaseHead:
    def test(self, constants, logging_setup, output_directory):
        """
        Tests the sqlite implementation of XchemBeDatabase.
        """

        database_specification = {
            "type": "soakdb3_lib.databases.aiosqlite",
            "filename": "%s/soakdb3.sqlite" % (output_directory),
        }

        # Test direct SQL access to the database.
        DatabaseTesterHead().main(
            constants,
            database_specification,
            output_directory,
        )


# ----------------------------------------------------------------------------------------
class TestDatabaseBody:
    def test(self, constants, logging_setup, output_directory):
        """
        Tests the sqlite implementation of XchemBeDatabase.
        """

        database_specification = {
            "type": "soakdb3_lib.databases.aiosqlite",
            "filename": "%s/soakdb3.sqlite" % (output_directory),
        }

        # Test direct SQL access to the database.
        DatabaseTesterBody().main(
            constants,
            database_specification,
            output_directory,
        )


# ----------------------------------------------------------------------------------------
class TestDatabaseSqliteBackupRestore:
    def test(self, constants, logging_setup, output_directory):
        """
        Tests the sqlite implementation of XchemBeDatabase.
        """

        database_specification = {
            "type": "soakdb3_lib.databases.aiosqlite",
            "filename": "%s/soakdb3.sqlite" % (output_directory),
        }

        # Test direct SQL access to the database.
        DatabaseTesterBackupRestore().main(
            constants,
            database_specification,
            output_directory,
        )


# ----------------------------------------------------------------------------------------
class _BaseTester:
    """
    Provide asyncio loop and error checking over *Tester classes.
    """

    def main(self, constants, specification, output_directory):
        """
        This is the main program which calls the test using asyncio.
        """

        multiprocessing.current_process().name = "main"

        failure_message = None
        try:
            # Run main test in asyncio event loop.
            asyncio.run(
                self._main_coroutine(constants, specification, output_directory)
            )

        except Exception as exception:
            logger.exception(
                "unexpected exception in the test method", exc_info=exception
            )
            failure_message = str(exception)

        if failure_message is not None:
            pytest.fail(failure_message)


# ----------------------------------------------------------------------------------------
class DatabaseTesterHead(_BaseTester):
    """
    Test direct SQL access to the database.
    """

    async def _main_coroutine(
        self, constants, database_specification, output_directory
    ):
        """ """

        databases = Databases()
        database = databases.build_object(database_specification)

        # Connect to database.
        await database.connect()

        try:
            # Write one record.
            await database.insert(
                Tablenames.HEAD,
                [{HeadFieldnames.LabVisit: "x", HeadFieldnames.Protein: "y"}],
            )

            all_sql = "SELECT * FROM soakDB"
            records = await database.query(all_sql)
            assert len(records) == 1
        finally:
            # Connect from the database... necessary to allow asyncio loop to exit.
            await database.disconnect()


# ----------------------------------------------------------------------------------------
class DatabaseTesterBody(_BaseTester):
    """
    Test direct SQL access to the database.
    """

    async def _main_coroutine(
        self, constants, database_specification, output_directory
    ):
        """ """

        databases = Databases()
        database = databases.build_object(database_specification)

        try:
            # Connect to database.
            await database.connect()

            uuid1 = 1000
            uuid2 = 2000
            uuid3 = 3000

            # Write one record.
            await database.insert(
                Tablenames.BODY,
                [{BodyFieldnames.LabVisit: "x", BodyFieldnames.ID: uuid1}],
            )
            all_sql = f"SELECT * FROM {Tablenames.BODY} ORDER BY ID ASC"
            records = await database.query(all_sql)
            assert len(records) == 1, "first %s count" % (all_sql)

            # Write two more records.
            await database.insert(
                Tablenames.BODY,
                [
                    {BodyFieldnames.LabVisit: "y", BodyFieldnames.ID: uuid2},
                    {BodyFieldnames.LabVisit: "z", BodyFieldnames.ID: uuid3},
                ],
            )
            records = await database.query(all_sql)
            assert len(records) == 3, "second %s count" % (all_sql)

            # Update one record to BUSY.
            await database.update(
                Tablenames.BODY, {BodyFieldnames.LabVisit: "z2"}, f"ID = {uuid3}"
            )
            z2_sql = f"SELECT * FROM {Tablenames.BODY} WHERE {BodyFieldnames.LabVisit} = 'z2' ORDER BY ID ASC"
            records = await database.query(z2_sql)
            assert len(records) == 1, "%s count" % z2_sql

            # Update two records to DEAD.
            await database.update(
                Tablenames.BODY,
                {BodyFieldnames.LabVisit: "u2"},
                f"ID IN ({uuid1}, {uuid2})",
            )
            u2_sql = f"SELECT * FROM {Tablenames.BODY} WHERE {BodyFieldnames.LabVisit} = 'u2' ORDER BY ID ASC"
            records = await database.query(u2_sql)
            assert len(records) == 2, "%s count" % u2_sql

        finally:
            # Connect from the database... necessary to allow asyncio loop to exit.
            await database.disconnect()


# ----------------------------------------------------------------------------------------
class DatabaseTesterBackupRestore(_BaseTester):
    """
    Test direct SQL backup and restore.
    """

    async def _main_coroutine(
        self, constants, database_specification, output_directory
    ):
        """ """

        databases = Databases()
        database = databases.build_object(database_specification)

        # Connect to database.
        await database.connect()

        try:
            uuid1 = 1000
            uuid2 = 2000

            # Write one record.
            await database.insert(
                Tablenames.BODY,
                [{BodyFieldnames.LabVisit: "x", BodyFieldnames.ID: uuid1}],
            )

            # Backup.
            await database.backup()

            # Write another record.
            await database.insert(
                Tablenames.BODY,
                [{BodyFieldnames.LabVisit: "y", BodyFieldnames.ID: uuid2}],
            )

            # Backup again (with two records)
            await database.backup()

            # Restore one in the past (when it had a single record).
            await database.restore(1)

            all_sql = (
                f"SELECT * FROM {Tablenames.BODY} ORDER BY ID ASC /* first query */"
            )
            records = await database.query(all_sql)
            assert len(records) == 1, "first %s count expected 1" % (all_sql)

            # Restore most recent (two records).
            await database.restore(0)

            all_sql = f"SELECT * FROM {Tablenames.BODY} ORDER BY ID ASC"
            records = await database.query(all_sql)
            assert len(records) == 2, "second %s count expected 2" % (all_sql)

        finally:
            # Connect from the database... necessary to allow asyncio loop to exit.
            await database.disconnect()



import asyncio
import logging
import multiprocessing
import os

import pytest
from dls_multiconf_lib.constants import ThingTypes as MulticonfThingTypes

# Configurator.
from dls_multiconf_lib.multiconfs import Multiconfs, multiconfs_set_default

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class BaseContextTester:
    """
    This is a base class for tests which use XchemBeContext.
    """

    def __init__(self):
        # For a capturing the news events as they arrive.
        self.consumed_news = []
        self.tasks_execution_outputs = {}
        self.residuals = ["stdout.txt", "stderr.txt", "main.log"]

    def main(self, constants, configuration_file, output_directory):
        """
        This is the main program which calls the test using asyncio.
        """

        # Save these for when the configuration is loaded.
        self.__configuration_file = configuration_file
        self.__output_directory = output_directory

        multiprocessing.current_process().name = "main"

        # self.__blocked_event = asyncio.Event()

        failure_message = None
        try:
            # Run main test in asyncio event loop.
            asyncio.run(self._main_coroutine(constants, output_directory))

        except Exception as exception:
            logger.exception(
                "unexpected exception in the test method", exc_info=exception
            )
            failure_message = str(exception)

        if failure_message is not None:
            pytest.fail(failure_message)

    # ----------------------------------------------------------------------------------------
    def get_multiconf(self):

        multiconf = Multiconfs().build_object(
            {
                "type": MulticonfThingTypes.YAML,
                "type_specific_tbd": {"filename": self.__configuration_file},
            }
        )

        # For convenience, always do these replacement.
        multiconf.substitute({"output_directory": self.__output_directory})

        # Add various things from the environment into the multiconf.
        multiconf.substitute(
            {
                "CWD": os.getcwd(),
                "PYTHONPATH": os.environ.get("PYTHONPATH", "PYTHONPATH"),
            }
        )

        # Set the global value of our multiconf which might be used in other modules.
        multiconfs_set_default(multiconf)

        return multiconf



import csv
import logging

# Client context creator.
from soakdb3_api.datafaces.context import Context as Soakdb3DatafaceClientContext
from soakdb3_api.datafaces.datafaces import datafaces_get_default

# Server context creator.
from soakdb3_lib.datafaces.context import Context as Soakdb3DatafaceServerContext

# Base class for the tester.
from tests.base_context_tester import BaseContextTester

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestMoveToDone:
    def test(self, constants, logging_setup, output_directory):
        """ """

        configuration_file = "tests/configurations/services.yaml"
        MoveToDoneTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class MoveToDoneTester(BaseContextTester):
    """
    Class to test the write_csv API call through the server.
    """

    async def _main_coroutine(self, constants, output_directory):
        """ """

        multiconf = self.get_multiconf()

        multiconf_dict = await multiconf.load()

        # Reference the dict entry for the soakdb3 dataface.
        soakdb3_dataface_specification = multiconf_dict[
            "soakdb3_dataface_specification"
        ]

        # Make the server context.
        soakdb3_server_context = Soakdb3DatafaceServerContext(
            soakdb3_dataface_specification
        )

        # Make the client context.
        soakdb3_client_context = Soakdb3DatafaceClientContext(
            soakdb3_dataface_specification
        )

        # Start the soakdb3 server context which includes the direct or network-addressable service.
        async with soakdb3_server_context:
            # Start the matching soakdb3 client context.
            async with soakdb3_client_context:
                await self.__run_the_csv_test(constants, output_directory)

    # ----------------------------------------------------------------------------------------

    async def __run_the_csv_test(self, constants, output_directory):
        """ """

        # The visitid to match what is in the test configuration yaml.
        visitid = output_directory

        dataface = datafaces_get_default()

        data = [
            ["a", "1"],
            ["b", "2"],
            ["c", '"3"'],
            ["d", "4,5,6"],
            ["e", "line1\nline2"],
            ["f", ""],
            ["g"],
            ["h", "done"],
        ]

        filename = "Batch-1_9bg5_2023-05-04_RI1000-0080-3drop.csv"

        # Write csv.
        await dataface.write_csv(visitid, data, f"echo/{filename}")

        # Move to done.
        transfer_type = "soak"
        transfer_subdirectory = "echo"
        await dataface.move_to_done(visitid, filename, transfer_type)

        # Read csv from the done.
        fieldnames = [0, 1]
        restval = "none"

        # Transfer type "soak" is considered "echo".
        with open(
            f"{output_directory}/lab36/{transfer_subdirectory}/done/{filename}"
        ) as stream:
            reader = csv.DictReader(stream, fieldnames=fieldnames, restval=restval)
            i = 0
            for row in reader:
                for col, value in row.items():
                    # Data row has at least this many columns?
                    if col < len(data[i]):
                        assert value == data[i][col], f"line {i} column {col}"
                    else:
                        assert value == restval, f"line {i} column {col}"
                i = i + 1



import logging

# Class for an aiohttp client.
from soakdb3_api.aiohttp_client import AiohttpClient

# XchemBeDataface protocolj things.
from soakdb3_api.datafaces.constants import Commands, Keywords

logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------------------
class Aiohttp:
    """
    Object implementing client side API for talking to the dataface server.
    Please see doctopic [A01].
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification=None):
        self.__specification = specification

        self.__aiohttp_client = AiohttpClient(
            specification["type_specific_tbd"]["aiohttp_specification"],
        )

    # ----------------------------------------------------------------------------------------
    def specification(self):
        return self.__specification

    # ----------------------------------------------------------------------------------------
    async def backup(self, visitid):
        """"""
        return await self.__send_protocolj("backup", visitid)

    # ----------------------------------------------------------------------------------------
    async def restore(self, visitid, nth):
        """"""
        return await self.__send_protocolj("restore", visitid, nth)

    # ----------------------------------------------------------------------------------------
    async def query(self, visitid, sql, subs=None, why=None):
        """"""
        return await self.__send_protocolj(
            "query",
            visitid,
            sql,
            subs=subs,
            why=why,
        )

    # ----------------------------------------------------------------------------------------
    async def query_for_dictionary(self, visitid, sql, subs=None, why=None):
        """"""
        return await self.__send_protocolj(
            "query_for_dictionary",
            visitid,
            sql,
            subs=subs,
            why=why,
        )

    # ----------------------------------------------------------------------------------------
    async def execute(self, visitid, sql, subs=None, why=None):
        """"""
        return await self.__send_protocolj(
            "execute",
            visitid,
            sql,
            subs=subs,
            why=why,
        )

    # ----------------------------------------------------------------------------------------
    async def insert(self, visitid, table_name, records, subs=None, why=None):
        """"""
        return await self.__send_protocolj(
            "insert",
            visitid,
            table_name,
            records,
            subs=subs,
            why=why,
        )

    # ----------------------------------------------------------------------------------------
    async def update(self, visitid, table_name, record, where, subs=None, why=None):
        """"""
        return await self.__send_protocolj(
            "update",
            visitid,
            table_name,
            record,
            where,
            subs=subs,
            why=why,
        )

    # ----------------------------------------------------------------------------------------
    async def update_head_fields(self, visitid, fields):
        """
        Handle update request from a soakdb range change.
        """

        return await self.__send_protocolj(
            "update_head_fields",
            visitid,
            fields,
        )

    # ----------------------------------------------------------------------------------------
    async def update_body_fields(self, visitid, fields):
        """
        Handle update request from a soakdb range change.
        """

        return await self.__send_protocolj(
            "update_body_fields",
            visitid,
            fields,
        )

    # ----------------------------------------------------------------------------------------
    async def write_csv(self, visitid, rows, filename):
        """
        Handle request to write rows as csv file.

        The rows are assumed fully prepared ahead of time.

        The filename's path should be a subdirectory (no leading slash).

        """

        return await self.__send_protocolj(
            "write_csv",
            visitid,
            rows,
            filename,
        )

    # ----------------------------------------------------------------------------------------
    async def move_to_done(
        self,
        visitid: str,
        csv_file: str,
        transfer_type: str,
    ) -> None:
        """
        Handle request to move a csv file to the done subdirectory.

        Args:
            visitid (str): full path to visit, including /processing
            csv_file (str): name of csv file with (no directory part included)
            transfer_type (str): either "soak", "cryo" or "shifter"
        """

        return await self.__send_protocolj(
            "move_to_done",
            visitid,
            csv_file,
            transfer_type,
        )

    # ----------------------------------------------------------------------------------------
    async def report_health(self):
        """"""
        return await self.__send_protocolj("report_health")

    # ----------------------------------------------------------------------------------------
    async def __send_protocolj(self, function, *args, **kwargs):
        """"""

        return await self.__aiohttp_client.client_protocolj(
            {
                Keywords.COMMAND: Commands.EXECUTE,
                Keywords.PAYLOAD: {
                    "function": function,
                    "args": args,
                    "kwargs": kwargs,
                },
            },
        )

    # ----------------------------------------------------------------------------------------
    async def close_client_session(self):
        """"""

        if self.__aiohttp_client is not None:
            await self.__aiohttp_client.close_client_session()

    # ----------------------------------------------------------------------------------------
    async def client_report_health(self):
        """"""

        return await self.__aiohttp_client.client_report_health()



import asyncio
import logging
import multiprocessing

import pytest

logger = logging.getLogger(__name__)


class Base:

    # ----------------------------------------------------------------------------------------
    def main(self, constants, infrastrcuture_context, output_directory):
        """ """

        multiprocessing.current_process().name = "main"

        failure_message = None
        try:
            # Run main test in asyncio event loop.
            asyncio.run(
                self._main_coroutine(
                    constants, infrastrcuture_context, output_directory
                )
            )

        except Exception as exception:
            logger.exception(
                "unexpected exception in the test method", exc_info=exception
            )
            failure_message = str(exception)

        if failure_message is not None:
            pytest.fail(failure_message)



import logging

from dls_utilpack.callsign import callsign
from dls_utilpack.thing import Thing

# Things created in the context.
from soakdb3_lib.datafaces.datafaces import Datafaces

logger = logging.getLogger(__name__)


thing_type = "soakdb3_lib.datafaces.context"


class Context(Thing):
    """
    Asyncio context for a dataface server object.
    On entering, it creates the object according to the specification (a dict).
    If configured, it starts the server as a coroutine, thread or process.
    On exiting, it commands the server to shut down.

    The enter and exit methods are exposed for use during testing.

    TODO: Add unit test for soakdb3_lib.datafaces.context.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification, predefined_uuid=None):
        Thing.__init__(self, thing_type, specification, predefined_uuid=predefined_uuid)

        # Reference to object which is a server, such as BaseAiohttp.
        self.server = None

        self.context_specification = self.specification().get("context", {})

    # ----------------------------------------------------------------------------------------
    async def aenter(self):
        """ """

        # Build the object according to the specification.
        self.server = Datafaces().build_object(self.specification())

        if self.context_specification.get("start_as") == "coro":
            await self.server.activate_coro()

        elif self.context_specification.get("start_as") == "thread":
            await self.server.start_thread()

        elif self.context_specification.get("start_as") == "process":
            await self.server.start_process()

    # ----------------------------------------------------------------------------------------
    async def aexit(self):
        """ """

        if self.server is not None:
            if self.context_specification.get("start_as") == "process":
                # Put in request to shutdown the server.
                await self.server.client_shutdown()

            if self.context_specification.get("start_as") == "coro":
                await self.server.direct_shutdown()

    # ----------------------------------------------------------------------------------------
    async def __aenter__(self):
        """ """

        await self.aenter()

    # ----------------------------------------------------------------------------------------
    async def __aexit__(self, type, value, traceback):
        """ """

        await self.aexit()

    # ----------------------------------------------------------------------------------------
    async def is_process_started(self):
        """"""

        if self.server is None:
            raise RuntimeError(f"{callsign(self)} a process has not been defined")

        try:
            return await self.server.is_process_started()
        except Exception:
            raise RuntimeError(
                f"unable to determing process started for server {callsign(self.server)}"
            )

    # ----------------------------------------------------------------------------------------
    async def is_process_alive(self):
        """"""

        if self.server is None:
            raise RuntimeError(f"{callsign(self)} a process has not been defined")

        try:
            return await self.server.is_process_alive()
        except Exception:
            raise RuntimeError(
                f"unable to determing dead or alive for server {callsign(self.server)}"
            )



from typing import Optional

from pydantic import BaseModel


class CrystalWellModel(BaseModel):
    """
    Model representing enough fields to inject desired crystal wells into the soadkb3 Main table.

    Typically this structure is used to inject new wells and their echo locations to soakdb3.

    ID is ignored when appending crystal wells, but populated by a fetch.
    """

    ID: Optional[str] = None
    LabVisit: Optional[str] = None
    CrystalPlate: Optional[str] = None
    CrystalWell: Optional[str] = None
    EchoX: Optional[int] = None
    EchoY: Optional[int] = None
    ProteinName: Optional[str] = None
    DropVolume: Optional[float] = None



import logging

import pytest

from soakdb3_api.databases.constants import BodyFieldnames, HeadFieldnames, Tablenames

# Client context creator.
from soakdb3_api.datafaces.context import Context as Soakdb3DatafaceClientContext
from soakdb3_api.datafaces.datafaces import datafaces_get_default

# Server context creator.
from soakdb3_lib.datafaces.context import Context as Soakdb3DatafaceServerContext

# Base class for the tester.
from tests.base_context_tester import BaseContextTester

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestDatafaceHead:
    def test(self, constants, logging_setup, output_directory):
        """ """

        configuration_file = "tests/configurations/services.yaml"
        DatafaceHeadTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class TestDatafaceBody:
    def test(self, constants, logging_setup, output_directory):
        """ """

        configuration_file = "tests/configurations/services.yaml"
        DatafaceBodyTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class DatafaceHeadTester(BaseContextTester):
    """
    Class to test the news.
    """

    async def _main_coroutine(self, constants, output_directory):
        """ """

        multiconf = self.get_multiconf()

        multiconf_dict = await multiconf.load()

        # Reference the dict entry for the soakdb3 dataface.
        soakdb3_dataface_specification = multiconf_dict[
            "soakdb3_dataface_specification"
        ]

        # Make the server context.
        soakdb3_server_context = Soakdb3DatafaceServerContext(
            soakdb3_dataface_specification
        )

        # Make the client context.
        soakdb3_client_context = Soakdb3DatafaceClientContext(
            soakdb3_dataface_specification
        )

        # Start the soakdb3 server context which includes the direct or network-addressable service.
        async with soakdb3_server_context:
            # Start the matching soakdb3 client context.
            async with soakdb3_client_context:
                await self.__run_the_head_test(constants, output_directory)

    # ----------------------------------------------------------------------------------------

    async def __run_the_head_test(self, constants, output_directory):
        """ """

        # The visitid to match what is in the test configuration yaml.
        visitid = output_directory

        dataface = datafaces_get_default()

        # Write one record.
        await dataface.insert(
            visitid,
            Tablenames.HEAD,
            [
                {
                    HeadFieldnames.LabVisit: "x",
                    HeadFieldnames.Protein: "y",
                }
            ],
        )

        # --------------------------------------------------------------------
        # Update like XChem soakdb range changed.
        fields = [
            {"field": HeadFieldnames.LabVisit, "value": "a"},
            {"field": HeadFieldnames.Protein, "value": "b"},
            {"field": HeadFieldnames.DropVolume, "value": 1.23},
        ]

        await dataface.update_head_fields(visitid, fields)

        z3_sql = "SELECT * from soakDB"
        records = await dataface.query_for_dictionary(visitid, z3_sql)
        assert len(records) == 1
        assert records[0][HeadFieldnames.LabVisit] == "a"
        assert records[0][HeadFieldnames.Protein] == "b"
        assert records[0][HeadFieldnames.DropVolume] == 1.23

        # Query for invalid visit.
        with pytest.raises(RuntimeError) as exception_info:
            records = await dataface.query_for_dictionary("badvisit", z3_sql)
        assert "badvisit" in str(exception_info.value)


# ----------------------------------------------------------------------------------------
class DatafaceBodyTester(BaseContextTester):
    """
    Class to test the news.
    """

    async def _main_coroutine(self, constants, output_directory):
        """ """

        multiconf = self.get_multiconf()

        multiconf_dict = await multiconf.load()

        # Reference the dict entry for the soakdb3 dataface.
        soakdb3_dataface_specification = multiconf_dict[
            "soakdb3_dataface_specification"
        ]

        # Make the server context.
        soakdb3_server_context = Soakdb3DatafaceServerContext(
            soakdb3_dataface_specification
        )

        # Make the client context.
        soakdb3_client_context = Soakdb3DatafaceClientContext(
            soakdb3_dataface_specification
        )

        # Start the soakdb3 server context which includes the direct or network-addressable service.
        async with soakdb3_server_context:
            # Start the matching soakdb3 client context.
            async with soakdb3_client_context:
                await self.__run_the_body_test(constants, output_directory)

    # ----------------------------------------------------------------------------------------

    async def __run_the_body_test(self, constants, output_directory):
        """ """

        # The visitid to match what is in the test configuration yaml.
        visitid = output_directory

        uuid1 = 1000
        uuid2 = 2000
        uuid3 = 3000

        dataface = datafaces_get_default()

        # Write one record.
        await dataface.insert(
            visitid,
            Tablenames.BODY,
            [
                {
                    BodyFieldnames.LabVisit: "x",
                    BodyFieldnames.ID: uuid1,
                }
            ],
        )

        all_sql = f"SELECT * FROM {Tablenames.BODY} ORDER BY ID ASC"
        records = await dataface.query(visitid, all_sql)
        assert len(records) == 2, "first %s count" % (all_sql)

        # Write two more records.
        await dataface.insert(
            visitid,
            Tablenames.BODY,
            [
                {
                    BodyFieldnames.LabVisit: "y",
                    BodyFieldnames.ID: uuid2,
                },
                {
                    BodyFieldnames.LabVisit: "z",
                    BodyFieldnames.ID: uuid3,
                },
            ],
        )
        records = await dataface.query(visitid, all_sql)
        assert len(records) == 4, "second %s count" % (all_sql)

        # Update one record.
        await dataface.update(
            visitid,
            Tablenames.BODY,
            {BodyFieldnames.LabVisit: "z2"},
            f"ID = '{uuid3}'",
        )
        z2_sql = f"SELECT * FROM {Tablenames.BODY} WHERE {BodyFieldnames.LabVisit} = 'z2' ORDER BY ID ASC"
        records = await dataface.query(visitid, z2_sql)
        assert len(records) == 2, "third %s count" % z2_sql

        # --------------------------------------------------------------------
        # Update like XChem soakdb range changed.
        fields = [
            {"id": uuid1, "field": BodyFieldnames.CompoundCode, "value": 15},
            {"id": uuid2, "field": BodyFieldnames.CompoundCode, "value": 20},
            {"id": uuid3, "field": BodyFieldnames.CompoundCode, "value": 20},
        ]

        await dataface.update_body_fields(visitid, fields)

        z3_sql = (
            f"SELECT ID, {BodyFieldnames.CompoundCode} FROM mainTable ORDER BY ID ASC"
        )
        records = await dataface.query(visitid, z3_sql)
        assert len(records) == 4, "fourth %s count" % z3_sql
        assert records[1][1] == "15", "first compound"
        assert records[2][1] == "20", "second compound"
        assert records[3][1] == "20", "third compound"

        # # Update two records to DEAD.
        # await database.xchem_be_body_table.update(
        #     {BodyFieldnames.LabVisit: "u2"}, f"ID IN ({uuid1}, {uuid2})"
        # )
        # u2_sql = (
        #     f"SELECT * FROM mainTable WHERE {BodyFieldnames.LabVisit} = 'u2' ORDER BY ID ASC"
        # )
        # records = await database.query(u2_sql)
        # assert len(records) == 2, "%s count" % u2_sql



from dls_servbase_api.aiohttp_client import AiohttpClient as DlsServbaseAiohttpClient


# ------------------------------------------------------------------------------------------
class AiohttpClient(DlsServbaseAiohttpClient):
    """
    Object representing a client which makes aiohttp requests.
    """

    pass



class NotFound(RuntimeError):
    pass


# When something has no value set yet.
class NotSet(RuntimeError):
    pass


class RemoteSubmitFailed(RuntimeError):
    pass


class CapacityReached(RuntimeError):
    pass


class DuplicateLabelException(RuntimeError):
    pass


class DuplicateUuidException(RuntimeError):
    pass


class DlsXchemBeflowClientConnectorError(RuntimeError):
    pass


class Factory:
    def build(qualname):
        if qualname == "soakdb3_api.exceptions.CapacityReached":
            return CapacityReached
        return None



import logging

# Base class for table definitions.
from dls_normsql.table_definition import TableDefinition

from soakdb3_api.databases.constants import (
    BodyFieldnames,
    HeadFieldnames,
    Tablenames,
    VisitFieldnames,
)

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class HeadTable(TableDefinition):
    # ----------------------------------------------------------------------------------------
    def __init__(self):
        TableDefinition.__init__(self, Tablenames.HEAD)

        self.fields[HeadFieldnames.Version] = {"type": "REAL"}
        self.fields[HeadFieldnames.LabVisit] = {"type": "TEXT"}
        self.fields[HeadFieldnames.Path] = {"type": "TEXT"}
        self.fields[HeadFieldnames.Protein] = {"type": "TEXT"}
        self.fields[HeadFieldnames.DropVolume] = {"type": "REAL"}
        self.fields[HeadFieldnames.CrystalsPerBatch] = {"type": "INTEGER"}
        self.fields[HeadFieldnames.OneBatchPerPlate] = {"type": "TEXT"}
        self.fields[HeadFieldnames.CompoundStock] = {"type": "REAL"}
        self.fields[HeadFieldnames.SolventPercent] = {"type": "REAL"}
        self.fields[HeadFieldnames.CryoStock] = {"type": "REAL"}
        self.fields[HeadFieldnames.DesiredCryo] = {"type": "REAL"}
        self.fields[HeadFieldnames.CryoLocation] = {"type": "TEXT"}
        self.fields[HeadFieldnames.DesiredSoakTime] = {"type": "REAL"}
        self.fields[HeadFieldnames.CrystalStartNumber] = {"type": "INTEGER"}
        self.fields[HeadFieldnames.BeamlineVisit] = {"type": "TEXT"}
        self.fields[HeadFieldnames.SpaceGroup] = {"type": "TEXT"}
        self.fields[HeadFieldnames.A] = {"type": "REAL"}
        self.fields[HeadFieldnames.B] = {"type": "REAL"}
        self.fields[HeadFieldnames.C] = {"type": "REAL"}
        self.fields[HeadFieldnames.alpha] = {"type": "REAL"}
        self.fields[HeadFieldnames.beta] = {"type": "REAL"}
        self.fields[HeadFieldnames.gamma] = {"type": "REAL"}
        self.fields[HeadFieldnames.Recipe] = {"type": "TEXT"}
        self.fields[HeadFieldnames.Resolution] = {"type": "REAL"}
        self.fields[HeadFieldnames.CentringMethod] = {"type": "TEXT"}
        self.fields[HeadFieldnames.EUBOpen] = {"type": "TEXT"}
        self.fields[HeadFieldnames.iNEXT] = {"type": "TEXT"}
        self.fields[HeadFieldnames.Covid19] = {"type": "TEXT"}
        self.fields[HeadFieldnames.ILOXchem] = {"type": "TEXT"}


# ----------------------------------------------------------------------------------------
class BodyTable(TableDefinition):
    # ----------------------------------------------------------------------------------------
    def __init__(self):
        TableDefinition.__init__(self, Tablenames.BODY)

        self.fields[BodyFieldnames.ID] = {
            "type": "INTEGER PRIMARY KEY",
            "index": True,
        }
        self.fields[BodyFieldnames.LabVisit] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.LibraryPlate] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.SourceWell] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.LibraryName] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.CompoundSMILES] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.CompoundCode] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.CrystalPlate] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.CrystalWell] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.EchoX] = {"type": "INTEGER"}
        self.fields[BodyFieldnames.EchoY] = {"type": "INTEGER"}
        self.fields[BodyFieldnames.DropVolume] = {"type": "REAL"}
        self.fields[BodyFieldnames.ProteinName] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.BatchNumber] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.CompoundStockConcentration] = {"type": "REAL"}
        self.fields[BodyFieldnames.CompoundConcentration] = {"type": "REAL"}
        self.fields[BodyFieldnames.SolventFraction] = {"type": "REAL"}
        self.fields[BodyFieldnames.SoakTransferVol] = {"type": "REAL"}
        self.fields[BodyFieldnames.SoakStatus] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.SoakTimestamp] = {"type": "UNKNOWN"}
        self.fields[BodyFieldnames.CryoStockFraction] = {"type": "REAL"}
        self.fields[BodyFieldnames.CryoFraction] = {"type": "REAL"}
        self.fields[BodyFieldnames.CryoWell] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.CryoTransferVolume] = {"type": "REAL"}
        self.fields[BodyFieldnames.CryoStatus] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.CryoTimestamp] = {"type": "UNKNOWN"}
        self.fields[BodyFieldnames.SoakingTime] = {"type": "REAL"}
        self.fields[BodyFieldnames.HarvestStatus] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.CrystalName] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.Puck] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.PuckPosition] = {"type": "INTEGER"}
        self.fields[BodyFieldnames.PinBarcode] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.MountingResult] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.MountingArrivalTime] = {"type": "UNKNOWN"}
        self.fields[BodyFieldnames.MountedTimestamp] = {"type": "UNKNOWN"}
        self.fields[BodyFieldnames.MountingTime] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.ispybStatus] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.DataCollectionVisit] = {"type": "VARCHAR"}
        self.fields[BodyFieldnames.SoakDBComments] = {"type": "TEXT"}


# ----------------------------------------------------------------------------------------
class VisitTable(TableDefinition):
    # ----------------------------------------------------------------------------------------
    def __init__(self):
        TableDefinition.__init__(self, Tablenames.VISIT)

        self.fields[VisitFieldnames.VISITID] = {"type": "VARCHAR"}
